{+--------------------------------------------------------------------------+
 | Component: TGifImage & TGif
 | Created: 4/12/96 11:52:35 PM
 | Author: Jeff Kinzer
 | Copyright 1996, all rights reserved.
 | Description: Displays / Converts GIF Files
 | Version: 2.0
 | Revision: 4.2
 | History:
 |    * Borland Pascal      -  GifUtl.pas     Sean Wenzel       ( CompuServe: 71736,1245 )
 |    * Created TGifImage   -  GifImg.pas     Jeff Kinzer       ( CompuServe: 102413,3557 )
 |    * Created TGif        -  Gif.pas        Jeff Kinzer
 | Unsupported GIF Features:
 |    * Misc. Extension Blocks:
 |        + Graphic Control extension
 |        + Comment extension
 |        + Plain text extension
 |        + Application extension
 | Note: I'm not sure what will happen if these
 |       blocks are encountered but it'll be interesting
 +--------------------------------------------------------------------------+}

unit Gif;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Menus;

const
{ General Purpose Constants }
  MAXSCREENWIDTH = 800;
  MAXCODES = 4095;				                { the maximum number of different codes 0 inclusive }

type
  TGifStatus = 0..100;

  TStatusChangeEvent = procedure(Sender: TObject; Status: TGifStatus) of object;

  PGifDataSubBlock = ^TGifDataSubBlock;
    TGifDataSubBlock = record
    Size: Byte;                                   { Size of the block -- 0 to 255 }
    Data: array[1..255] of Byte;                  { The data }
  end;

  PGifHeader = ^TGifHeader;
  TGifHeader = record
    Signature: array[0..2] of Char;               { Contains 'GIF' }
    Version: array[0..2] of Char;                 { '87a' or '89a' }
  end;

  TLogicalScreenDescriptor = record
    ScreenWidth: Word;                            { Logical screen width }
    ScreenHeight: Word;                           { Logical screen height }
    PackedFields: Byte;                           { Packed fields (See Below) }
    BackGroundColorIndex: Byte;                   { Index to global color table }
    AspectRatio: Byte;                            { Actual ratio = (AspectRatio + 15) / 64 }
  end;

  TColorItem = record			                        { One item color table }
    Red: Byte;
    Green: Byte;
    Blue: Byte;
  end;

  TColorTable = array[0..255] of TColorItem;	      { The color table }

  TGifImageDescriptor = record
    Separator: Byte;		                  { Fixed value of ImageSeparator }
    ImageLeftPos: Word; 	                { Column in respect to left edge of logical screen }
    ImageTopPos: Word;	                  { Row in respect to top of logical screen }
    ImageWidth: Word;		                  { Width of image in pixels }
    ImageHeight: Word; 	                  { Height of image in pixels }
    PackedFields: Byte;	                  { See below }
  end;

  TGifInfo = record
    Header: TGifHeader;
    LogicalScreen: TLogicalScreenDescriptor;
    ImageDescriptor: TGifImageDescriptor;
    Interlaced: Boolean;                            { True if image is interlaced }
    UseLocalColors: Boolean;		                    { True if local colors in use }
  end;

(*
  TExtensionBlock = record
    Introducer: Byte;                               { Fixed value of ExtensionIntroducer }
    ExtensionLabel: Byte;
    BlockSize: Byte;
  end;

  PGifCodeItem = ^TGifCodeItem;
  TGifCodeItem = record
    Code1,
    Code2: Byte;
  end;
*)

  TGraphicLine = array [0..2048] of Byte;

  TGifVersion = (v87a, v89a);

  PGif = ^TGif;
  TGif = class(TGraphic)
  private
    FGifVersion: TGifVersion;
    FOnChange: TNotifyEvent;
    FOnStatusChange: TStatusChangeEvent;
  public
    FBitmap: TBitmap;
  private
    { GIF Data }
    GifStream: TMemoryStream;	                  { the file stream for the gif file }
    GifInfo: TGifInfo;                          { gif file header }
    GlobalColorTable: TColorTable;		          { global color table }
    LocalColorTable: TColorTable;			          { local color table }
    LZWCodeSize: Byte;				                  { minimum size of the LZW codes in bits }
    ImageData: TGifDataSubBlock;		            { variable to store incoming gif data }
    TableSize: Word;					                  { number of entrys in the color table }
    BitsLeft,                                   { bits left in Byte  }
    BytesLeft: Integer;	                        { Bytes left in block }
    BadCodeCount: Word;          	              { bad code counter }
    CurrCodeSize: Integer;                     	{ Current size of code in bits }
    ClearCode: Integer;                        	{ Clear code value }
    EndingCode: Integer;                       	{ ending code value }
    Slot: Word;							                    { where next new code will be added }
    TopSlot: Word;		                        	{ highest slot position for the current code size }
    HighCode: Word;		                          { highest code that does not require decoding }
    NextByte: Integer;                        	{ the index to the next Byte in the datablock array }
    CurrByte: Byte;  	                        	{ the current Byte }
    DecodeStack: array[0..MAXCODES] of Byte;    { stack for the decoded codes }
    Prefix: array[0..MAXCODES] of Integer;      { array for code prefixes }
    Suffix: array[0..MAXCODES] of Integer;      { array for code suffixes }
    LineBuffer: TGraphicLine;                   { array for buffer line output }
    CurrentX,
    CurrentY: Integer;                          { current screen locations }
    InterlacePass: Byte;                        { interlace pass number }
    { Bitmap Data }
    BmHeader: TBitmapInfoHeader;                { File Header for bitmap file }
    BmpStream: TMemoryStream;
    ImageLines: TList;                          { Image data }
    { Status Variables }
    FStatus: TGifStatus;
    CurStatus,
    MaxStatus: Integer;
    FFileName: TFileName;
    { Member Functions }
    procedure GifError(ErrorStr: String);
    procedure DecompressGif;
    procedure ConvertGifToBmp(var ABitmap: TBitmap);
    procedure SetBitmap(newValue: TBitmap);
    function GetGifVersion: TGifVersion;
    procedure SetGifVersion(newValue: TGifVersion);
    function GetInterlaced: Boolean;
    procedure SetInterlaced(newValue: Boolean);
  protected
    procedure Draw(ACanvas: TCanvas; const Rect: TRect); override;
    function GetEmpty: Boolean; override;
    function GetHeight: Integer; override;
    function GetWidth: Integer; override;
    procedure ReadData(Stream: TStream); override;
    procedure SetHeight(Value: Integer); override;
    procedure SetWidth(Value: Integer); override;
    procedure WriteData(Stream: TStream); override;
    procedure ChangeStatus(LoadStatus: LongInt); virtual;
    procedure TriggerChangeEvent; virtual;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure BmpToGif(var ABitmap: TBitmap); virtual;
    procedure GifToBmp(var ABitmap: TBitmap); virtual;
    procedure Assign(Source: TPersistent); override;
    procedure LoadFromFile(const Filename: string); override;
    procedure SaveToFile(const Filename: string); override;
    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property Status: TGifStatus read FStatus;
    property FileName: TFileName read FFileName;
    property OnStatusChange: TStatusChangeEvent read FOnStatusChange write FOnStatusChange;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property GifVersion: TGifVersion read GetGifVersion write SetGifVersion stored True;
    property Interlaced: Boolean read GetInterlaced write SetInterlaced stored True;
  end;

  EGifException = class(Exception);

implementation

function Power(PowerNum, BaseNum: Real): Real;
begin
  Result := Exp(BaseNum * Ln(PowerNum));
end;

{ TGif }

type
  PBitmapLineStruct = ^TBitmapLineStruct;
  TBitmapLineStruct = record
    LineData: TGraphicLine;
    LineNo: Integer;
  end;

const
{ General Purpose Constants }
  BlockTerminator: Byte = 0;              { terminates stream of data blocks }
  Trailer: Byte = $3B;			              { indicates the end of the GIF data stream }
  ExtensionIntroducer: Byte = $21;
  ImageSeparator: Byte = $2C;

{ Image Descriptor Bit Masks }
  idLocalColorTable = $80;	              { set if a local color table follows }
  idInterlaced = $40;			                { set if image is interlaced }
  idSort = $20;					                  { set if color table is sorted }
  idReserved = $0C;				                { reserved - must be set to $00 }
  idColorTableSize = $07;		              { size of color table as above }

{ Logical Screen Descriptor Packed Field Masks }
  lsdGlobalColorTable = $80;	            { set if global color table follows L.S.D. }
  lsdColorResolution = $70;	              { Color resolution - 3 bits }
  lsdSort = $08;					                { set if global color table is sorted - 1 bit }
  lsdColorTableSize = $07;	              { size of global color table - 3 bits }
  									                      { Actual size = 2^value+1    - value is 3 bits }
{ Error Constants }
  ErrNoError = '';                                    { No errors found }
  ErrFileNotFound = 'GIF file was not found';         { Gif file not found }
  ErrNotGIF = 'Invalid GIF file format';               { File is not a gif file }
  ErrNoGlobalColor = 'GIF color table not found';     { No Global Color table found }
  ErrImagePreceded = 'Unrecognized data in GIF image descriptor';    { image descriptor preceeded by unknown data }
  ErrEmptyBlock = 'GIF image has no data'; 	          { Block has no data }
  ErrUnExpectedEOF = 'Unexpected end of GIF file';    { unexpected EOF }
  ErrBadCodeSize = 'Bad GIF code size';               { bad code size }
  ErrBadCode = 'Bad GIF code';   		                  { Bad code was found }
  ErrBitSizeOverflow = 'Bad GIF bit size';            { bit size went beyond 12 bits }
  ErrNoBMP = 'GIF is empty'; 			                    { Could not make BMP file }

{ Bit masks for use with NextCode }
  CodeMask: array[0..12] of Integer = ($0000, $0001, $0003, $0007, $000F, $001F, $003F,
                                       $007F, $00FF, $01FF, $03FF, $07FF, $0FFF);

function TGif.GetGifVersion: TGifVersion;
begin
  case Ord(GifInfo.Header.Version[1]) of
    55: Result := v87a;
    57: Result := v89a;
  end;
end;  { GetGifVersion }

procedure TGif.SetGifVersion(NewValue: TGifVersion);
begin
{
  if GifVersion <> NewValue then
  begin
    FGifVersion := NewValue;
  end;
}
end;  { SetGifVersion }

function TGif.GetInterlaced: Boolean;
begin
  Result := GifInfo.Interlaced;
end;  { GetInterlaced }

procedure TGif.SetInterlaced(newValue: Boolean);
begin
{
  if FInterlaced <> NewValue then
  begin
    FInterlaced := NewValue;
  end;
}
end;  { SetInterlaced }

procedure TGif.TriggerChangeEvent;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;  {TriggerChangeEvent}

procedure TGif.SetBitmap(NewValue: TBitmap);
begin
  if FBitmap <> NewValue then
  begin
    FBitmap.Assign(NewValue);
  end;
end;  {SetBitmap}

procedure TGif.Draw(ACanvas: TCanvas; const Rect: TRect);
begin
  StretchBlt(ACanvas.Handle, Rect.Left, Rect.Top, Rect.Right - Rect.Left,
      Rect.Bottom - Rect.Top, FBitmap.Canvas.Handle, 0, 0, Width, Height, ACanvas.CopyMode);
end;  {Draw}

function TGif.GetEmpty: Boolean;
begin
  Result := FBitmap.Empty;
end;  {GetEmpty}

function TGif.GetHeight: Integer;
begin
  Result := GifInfo.ImageDescriptor.ImageHeight;
end;  {GetHeight}

function TGif.GetWidth: Integer;
begin
  Result := GifInfo.ImageDescriptor.ImageWidth;
end;  {GetWidth}

procedure TGif.ReadData(Stream: TStream);
var
  AVersion: array[0..2] of Char;
begin
  if Assigned(Stream) then
  begin
    Stream.Seek(0, 0);
    Stream.Read(AVersion, 3);
    Stream.Seek(0, 0);
    if AVersion = 'GIF' then
      inherited ReadData(Stream)
    else
      FBitmap.LoadFromStream(Stream);
  end;
end;  {ReadData}

procedure TGif.SetHeight(Value: Integer);
begin
{  TriggerChangeEvent;}
end;  {SetHeight}

procedure TGif.SetWidth(Value: Integer);
begin
{  TriggerChangeEvent;}
end;  {SetWidth}

procedure TGif.WriteData(Stream: TStream);
begin
  inherited WriteData(Stream);
end;  {WriteData}

procedure TGif.Assign(Source: TPersistent);
begin
  if Source is TGif then
    FBitmap.Assign(TGif(Source).FBitmap)
  else
    inherited Assign(Source);
end;  {Assign}

procedure TGif.LoadFromFile(const Filename: string);
begin
  if FFilename <> FileName then
  begin
    FFileName := FileName;
    if (UpperCase(ExtractFileExt(FFilename)) = '.GIF') and FileExists(FFilename) then
    begin
      GifStream.LoadFromFile(FFilename);             { Load the file into memory }
      GifToBmp(FBitmap);
    end;
    TriggerChangeEvent;
  end;
end;  {LoadFromFile}

procedure TGif.SaveToFile(const Filename: string);
begin
(*
  if UpperCase(ExtractFileExt(Filename)) = '.BMP' then
  begin
    WriteBmpStream;
    BmpStream.SaveToFile(Filename);
  end
  else
{    Error}
*)
end;  {SaveToFile}

procedure TGif.LoadFromStream(Stream: TStream);
begin
  if Assigned(Stream) then
  begin
    GifStream.LoadFromStream(Stream);
    GifToBmp(FBitmap);
  end;
  TriggerChangeEvent;
end;  {LoadFromStream}

procedure TGif.SaveToStream(Stream: TStream);
var
  MemStream: TMemoryStream;
begin
  if Assigned(Stream) then
  begin
    MemStream := TMemoryStream.Create;
    ConvertGifToBmp(FBitmap);
    MemStream.LoadFromStream(BmpStream);
    MemStream.SaveToStream(Stream);
    MemStream.Free;
  end;
end;  {SaveToStream}

destructor TGif.Destroy;
begin
  if Assigned(FBitmap) then
    FBitmap.Free;
  if Assigned(GifStream) then
    GifStream.Free;
  if Assigned(BmpStream) then
    BmpStream.Free;
  if Assigned(ImageLines) then
    ImageLines.Free;
  inherited Destroy;
end;  {Destroy}

constructor TGif.Create;
begin
  inherited Create;
  FBitmap := TBitmap.Create;
  GifStream := TMemoryStream.Create;
  BmpStream := TMemoryStream.Create;
  ImageLines := TList.Create;
  FStatus := 0;
  FFileName := '';
end;  {Create}

procedure TGif.BmpToGif(var ABitmap: TBitmap);

  procedure CreateGifHeader;
  { This routine takes the values from the BMP
    image descriptor and fills in the appropriate
    values in the GIF struct }
{  var
    TmpField: Byte;}
  begin
    with GifInfo.Header do                                { Write GIF Header }
    begin
      Signature := 'GIF';                         { Set 'GIF' signature }
      Version := '8?a';                           { Fill in GIF version, with '?' for blank }
      case Ord(GetGifVersion) of                    { If FGifVersion = }
        0: Version[1] := '7';                       { v87a then replace '?' with '7' }
        1: Version[1] := '9';                       { v89a then replace '?' with '9' }
      end;
    end;
    with GifInfo.LogicalScreen do                         { Write GIF Logical Screen Descriptor }
    begin
      ScreenWidth := BmHeader.biWidth;
      ScreenHeight := BmHeader.biHeight;
{      TmpField := lsdGlobalColorTable;}
      PackedFields := lsdGlobalColorTable;
      BackGroundColorIndex := 0;
      AspectRatio := 0;
    end;
    with GifInfo.ImageDescriptor do
    begin
      Separator := ImageSeparator;
      ImageLeftPos := 0;
      ImageTopPos := 0;
      ImageWidth := BmHeader.biWidth;
      ImageHeight := BmHeader.biHeight;
      PackedFields := 0;                        { PackedFields:   ?????111 - 256 Colors }
      if GifInfo.UseLocalColors then                          {   ?????011 - 16 Colors  }
        PackedFields := PackedFields or idLocalColorTable;    {   ?????000 - 2 Colors   }
      if GifInfo.Interlaced then
        PackedFields := PackedFields or idInterlaced;
    end;
{    if GifInfo.UseLocalColors then
    begin
      LocalTable}
  end;

begin
end; { BmpToGif }

procedure TGif.GifToBmp(var ABitmap: TBitmap);

  procedure CreateBitmapHeader;
  { This routine takes the values from the GIF
    image descriptor and fills in the appropriate
    values in the bitmap header struct }
  begin
    with BmHeader do
    begin
      biSize := SizeOf(TBitmapInfoHeader);
      biWidth := GifInfo.ImageDescriptor.ImageWidth;
      biHeight := GifInfo.ImageDescriptor.ImageHeight;
      biPlanes := 1;                           { Arcane and rarely used }
      biBitCount := 8;                         { Should this be hardcoded? }
      biCompression := BI_RGB;                 { No compression in this version }
      biSizeImage := 0;                        { Because we are not compressing image }
      biXPelsPerMeter :=143;                   { Rarely used very arcane field }
      biYPelsPerMeter :=143;                   { Rarely used very arcane field }
      biClrUsed := 0;                          { all colors are used }
      biClrImportant := 0;                     { all colors are important }
    end;
  end;

  procedure DecodeGifHeader;
  { Decodes the header and palette info }
  var
    i: LongInt;
    p: Pointer;
  begin
    GifStream.Read(GifInfo.Header, SizeOf(GifInfo.Header));                     { read the header }
    if GifInfo.Header.Signature <> 'GIF' then
      GifError(ErrNotGif);                                          { is vaild signature? }
    GifStream.Read(GifInfo.LogicalScreen, SizeOf(GifInfo.LogicalScreen));       { Decode header information }
    if GifInfo.LogicalScreen.PackedFields and lsdGlobalColorTable = lsdGlobalColorTable then
    begin
      TableSize := Trunc(Power(2, (GifInfo.LogicalScreen.PackedFields and lsdColorTableSize) + 1));
      { Read Global Color Table }
      GifStream.Read(GlobalColorTable, TableSize * SizeOf(TColorItem));
    end
    else
      GifError(ErrNoGlobalColor);
    GifStream.Read(GifInfo.ImageDescriptor, SizeOf(GifInfo.ImageDescriptor));   { Read image descriptor }
    if GifInfo.ImageDescriptor.Separator <> ImageSeparator then         { Verify the descriptor }
      GifError(ErrImagePreceded);
    { Check for local color table }
    if GifInfo.ImageDescriptor.PackedFields and idLocalColorTable = idLocalColorTable then
    begin
      TableSize := Trunc(Power(2, (GifInfo.ImageDescriptor.PackedFields and idColorTableSize) + 1));
      { Read Local Color Table }
      GifStream.Read(LocalColorTable, TableSize * SizeOf(TColorItem));
      GifInfo.UseLocalColors := True;
    end
    else
      GifInfo.UseLocalColors := False;
    {Check for interlaced}
    GifInfo.Interlaced := GifInfo.ImageDescriptor.PackedFields and idInterlaced = idInterlaced;
    if GifInfo.Interlaced then
      InterlacePass := 0;
    { Reset then Expand capacity of the Image Lines list }
    for i := (ImageLines.Count - 1) downto 0 do
    begin
      p := ImageLines.Items[i];
      ImageLines.Delete(i);
      Dispose(p);
    end;
    ImageLines.Clear;
    ImageLines.Capacity := GifInfo.ImageDescriptor.ImageHeight;
    if (GifStream = nil) then
      GifError(ErrFileNotFound);
  end; { BuildGifInfo }

begin
  DecodeGifHeader;
  { Create the bitmap header info }
  MaxStatus := (GifInfo.ImageDescriptor.ImageHeight * 2);
  ChangeStatus(0);
  CreateBitmapHeader;
  { Decode the GIF }
  DecompressGif;
  ConvertGifToBmp(ABitmap);
end;

procedure TGif.DecompressGif;

  var
    SP: Integer;

  procedure DecodeLZWCode(var Code: Word);        { Local procedure that decodes a }
  begin                                             { code and puts it on the decode stack }
    while Code > HighCode do                      { rip thru the prefix list placing suffixes }
    begin                                         { onto the decode stack }
      DecodeStack[SP] := Suffix[Code];            { put the suffix on the decode stack }
      inc(SP);                                    { increment decode stack index }
      Code := Prefix[Code];                       { get the new prefix }
    end;
    DecodeStack[SP] := Code;        	            { put the last code onto the decode stack }
    inc(SP);                    			            { increment the decode stack index }
  end; { DecodeLZWCode }

  procedure InitCompressionStream;
  begin
    GifStream.Read(LZWCodeSize, SizeOf(Byte));	  { get minimum code size }
    if not (LZWCodeSize in [2..9]) then     	    { valid code sizes 2-9 bits }
       GifError(ErrBadCodeSize);
    CurrCodeSize := Succ(LZWCodeSize);            { set the initial code size }
    ClearCode := 1 shl LZWCodeSize;               { set the clear code }
    EndingCode := Succ(ClearCode);                { set the ending code }
    HighCode := Pred(ClearCode);   		            { set the highest code not needing decoding }
    BytesLeft := 0;                               { clear other variables }
    BitsLeft := 0;
    CurrentX := 0;
    CurrentY := 0;
  end; { InitCompressionStream }

  procedure ReadSubBlock;
  begin
    Application.ProcessMessages;
    GifStream.Read(ImageData.Size, SizeOf(ImageData.Size));   { get the data block size }
    if ImageData.Size = 0 then
      GifError(ErrEmptyBlock);									              { check for empty block }
    GifStream.Read(ImageData.Data, ImageData.Size); 	        { read in the block }
    NextByte := 1;                                            { reset next Byte }
    BytesLeft := ImageData.Size;        				              { reset Bytes left }
  end; { ReadSubBlock }

  function NextCode: Word;                        { returns a code of the proper bit size }
  begin
    if BitsLeft = 0 then    	                    { any bits left in Byte ? }
    begin 						                            { any Bytes left }
      if BytesLeft <= 0 then 	                    { if not get another block }
        ReadSubBlock;
      CurrByte := ImageData.Data[NextByte]; 	    { get a Byte }
      inc(NextByte);                        	    { set the next Byte index }
      BitsLeft := 8;                        	    { set bits left in the Byte }
      dec(BytesLeft);                       	    { decrement the Bytes left counter }
    end;
    Result := CurrByte shr (8 - BitsLeft);		  	{ shift off any previosly used bits}
    while CurrCodeSize > BitsLeft do        	    { need more bits ? }
    begin
      if BytesLeft <= 0 then						          { any Bytes left in block ? }
        ReadSubBlock;                             { if not read in another block }
      CurrByte := ImageData.Data[NextByte];      	{ get another Byte }
      inc(NextByte);                             	{ increment NextByte counter }
      Result := Result or
          (CurrByte shl BitsLeft);	              { add the remaining bits to the return value }
      BitsLeft := BitsLeft + 8;                   { set bit counter }
      dec(BytesLeft);                     	      { decrement Bytesleft counter }
    end;
    BitsLeft := BitsLeft - CurrCodeSize;          { subtract the code size from bitsleft }
    Result := Result and CodeMask[CurrCodeSize];  { mask off the right number of bits }
  end; { NextCode }

  procedure CreateLine;
  {fills in Line list with current line}
  var
    p: PBitmapLineStruct;
  begin
    New(p);                                       {Create a new bmp line}
    p^.LineData := LineBuffer;                    {Fill in the data}
    p^.LineNo := CurrentY;
    ImageLines.Add(p);                            { Add it to the list of lines }
    inc(CurrentY);                                { Prepare for the next line }
    ChangeStatus(CurStatus + 1);
    if GifInfo.Interlaced then                            { Interlace support }
    begin
      case InterlacePass of
        0: CurrentY := CurrentY + 7;
        1: CurrentY := CurrentY + 7;
        2: CurrentY := CurrentY + 3;
        3: CurrentY := CurrentY + 1;
        end;
      if CurrentY >= GifInfo.ImageDescriptor.ImageHeight then
      begin
        inc(InterLacePass);
        case InterLacePass of
          1: CurrentY := 4;
          2: CurrentY := 2;
          3: CurrentY := 1;
        end;
      end;
    end;
  end; { CreateLine }

var
  TempOldCode, OldCode: Word;
  BufCnt: Word;	                  { line buffer counter }
  Code, C: Word;
  CurrBuf: Word;	                { line buffer index }
  MaxVal: Boolean;
begin
  InitCompressionStream;                          { Initialize decoding paramaters }
  OldCode := 0;
  SP := 0;
  BufCnt := GifInfo.ImageDescriptor.ImageWidth;           { set the Image Width }
  CurrBuf := 0;
  MaxVal := False;
  C := NextCode;				                          { get initial code - should be clear code }
  while C <> EndingCode do                        { main loop until ending code is found }
  begin
    if C = ClearCode then	                        { code is clear code - so clear }
    begin
      CurrCodeSize := LZWCodeSize + 1;	          { reset the code size }
      Slot := EndingCode + 1;				              { set slot for next new code }
      TopSlot := 1 shl CurrCodeSize;	            { set max slot number }
      while C = ClearCode do
        C := NextCode;	                          { read until all clear codes gone }
      if C = EndingCode then
        GifError(ErrBadCode);   	                { ending code after a clear code }
      if C >= Slot then                           { if beyond preset codes then set to zero }
        C := 0;
      OldCode := C;
      DecodeStack[sp] := C; 	                    { output code to decoded stack }
      inc(SP);        			                      { increment decode stack index }
    end
    else                                          { not  clear code or ending code so must }
    begin                                         { be a code code - so decode the code }
      Code := C;
      if Code < Slot then                         { is the code in the table? }
      begin
        DecodeLZWCode(Code);                         { decode the code }
        if Slot <= TopSlot then
        begin               			                { add the new code to the table }
          Suffix[Slot] := Code;  		              { make the suffix }
          PreFix[slot] := OldCode; 	              { the previous code - a link to the data }
          inc(Slot);     	                        { increment slot number }
          OldCode := C;    	                      { set oldcode }
        end;
        if Slot >= TopSlot then 		              { have reached the top slot for bit size }
        begin                   	                { increment code bit size }
          if CurrCodeSize < 12 then 	            { new bit size not too big? }
          begin
            TopSlot := TopSlot shl 1;	            { new top slot }
            inc(CurrCodeSize) 			              { new code size }
          end
          else
            MaxVal := True; 			                { Must check next code is a start code }
        end;
      end
      else
      begin	                                      { the code is not in the table }
        if Code <> Slot then
          GifError(ErrBadCode);                   { so error out }
        { code does not exist, so make
          a new entry in the code table
          and then translate the new code }
        TempOldCode := OldCode;                   { make a copy of the old code }
        while OldCode > HighCode do 	            { translate the old code and place it }
        begin              			                  { on the decode stack }
          DecodeStack[SP] := Suffix[OldCode];     { do the suffix }
          OldCode := Prefix[OldCode];             { get next prefix }
        end;
        DecodeStack[SP] := OldCode;	              { put the code onto the decode stack
                                                    but DO NOT increment stack index }
        { the decode stack is not
          incremented because because
          we are only translating the oldcode
          to get the first Character }
        if Slot <= TopSlot then
        begin 	                                  { make new code entry }
          Suffix[Slot] := OldCode;     	          { first Char of old code }
          Prefix[Slot] := TempOldCode; 	          { link to the old code prefix }
          inc(Slot);                   	          { increment slot }
        end;
        if Slot >= TopSlot then                   { slot is too big }
        begin						                          { increment code size }
          if CurrCodeSize < 12 then
          begin
            TopSlot := TopSlot shl 1;	            { new top slot }
                inc(CurrCodeSize);   	            { new code size }
          end
          else
              MaxVal := True; 			              { Must check next code is a start code }
        end;
        DecodeLZWCode(Code);                         { now that the table entry exists decode it }
        OldCode := C;                             { set the new old code }
      end;
    end;
    { the decoded string is on the decode
      stack so pop it off and put it into
      the line buffer }
    while SP > 0 do
    begin
      dec(SP);
      LineBuffer[CurrBuf] := DecodeStack[SP];
      inc(CurrBuf);
      dec(BufCnt);
      if BufCnt = 0 then                          { is the line full ? }
      begin
        CreateLine;
        CurrBuf := 0;
        BufCnt := GifInfo.ImageDescriptor.ImageWidth;
      end;
    end;
    C := NextCode;	                              { get the next code and go at it some more }
    if (MaxVal = True) and (C <> ClearCode) then
      GifError(ErrBitSizeOverflow);
    MaxVal := False;
  end;
end; { GifDecode }

procedure TGif.ConvertGifToBmp(var ABitmap: TBitmap);
var
  BitFile: TBitmapFileHeader;
  i,
  Line,
  x: Integer;
  Ch: Char;
  p: PBitmapLineStruct;
begin
  BitFile.bfSize := (3 * 255) + SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader) +
      (GifInfo.ImageDescriptor.ImageHeight * GifInfo.ImageDescriptor.ImageWidth);
  BitFile.bfReserved1 := 0;                       { Not currently used }
  BitFile.bfReserved2 := 0;                       { Not currently used }
  BitFile.bfOffBits := (4 * 256) + SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader);
  BmpStream.Clear;                                { Write the file header }
{  BmpStream.Seek(0,0);}
  Ch := 'B';
  BmpStream.Write(Ch, 1);
  Ch := 'M';
  BmpStream.Write(Ch, 1);
  BmpStream.Write(BitFile.bfSize, SizeOf(BitFile.bfSize));
  BmpStream.Write(BitFile.bfReserved1, SizeOf(BitFile.bfReserved1));
  BmpStream.Write(BitFile.bfReserved2, SizeOf(BitFile.bfReserved2));
  BmpStream.Write(BitFile.bfOffBits, SizeOf(BitFile.bfOffBits));
  BmpStream.Write(BmHeader, SizeOf(BmHeader));    { Write the bitmap image header info }
  { Write the RGB palete
    inforamtion to this file }
  if GifInfo.UseLocalColors then                          { Use the local color table }
  begin
    for i:= 0 to 255 do
    begin
      BmpStream.Write(LocalColorTable[i].Blue, 1);
      BmpStream.Write(LocalColorTable[i].Green, 1);
      BmpStream.Write(LocalColorTable[i].Red, 1);
      BmpStream.Write(Ch, 1);                     { Bogus palete entry required by windows }
    end;
  end
  else                                            { Use the global table }
  begin
    for i:= 0 to 255 do
    begin
      BmpStream.Write(GlobalColorTable[i].Blue, 1);
      BmpStream.Write(GlobalColorTable[i].Green, 1);
      BmpStream.Write(GlobalColorTable[i].Red, 1);
      BmpStream.Write(Ch, 1);                     { Bogus palete entry required by windows }
    end;
  end;
  Line := GifInfo.ImageDescriptor.ImageHeight;            { Init the Line Counter }
  {Write out File lines in reverse order}
  while Line >= 0 do
  begin
    { Go through the line list in reverse
      order looking for the current Line.
      Use reverse order since non interlaced
      gifs are stored top to bottom.  Bmp file
      needs to be written bottom to top }
    for i := (ImageLines.Count - 1) downto 0  do
    begin
      ChangeStatus(CurStatus + 1);
      p := ImageLines.Items[i];
      if p^.LineNo = Line then
      begin
        x := GifInfo.ImageDescriptor.ImageWidth;
        BmpStream.Write(p^.LineData, x);
        Ch := chr(0);
        while (x and 3) <> 0 do                   { Pad up to 4-Byte boundary with zeroes }
        begin
          inc(x);
          BmpStream.Write(Ch, 1);
        end;
        Break;
      end;
    end;
    Dec(Line);
  end;
  BmpStream.Seek(0, 0);                           { Reset mewmory stream}
  ABitmap.LoadFromStream(BmpStream);
  ChangeStatus(0);
end; { WriteBmpStream }

procedure TGif.ChangeStatus(LoadStatus: LongInt);
begin
  if LoadStatus > MaxStatus then
  begin
    FStatus := 100;
    CurStatus := MaxStatus;
  end
  else
  begin
    FStatus := Round((LoadStatus / MaxStatus) * 100);
    CurStatus := LoadStatus;
  end;
  if Assigned(FOnStatusChange) then
    FOnStatusChange(Self, FStatus);
end;  {TriggerStatusChangeEvent}

procedure TGif.GifError(ErrorStr: String);
begin
  if ErrorStr > ' ' then
    raise EGifException.Create(ErrorStr);
end;

end.
